//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* file_io.cpp                                                            *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <map>
#include <iostream>

#include "mystd.h"
#include "olm.h"
#include "asi.h"
#include "file_io.h"
#include "yamc.h"

using namespace std;

namespace NFileIO {
    // actually, only FileI is implemented here;
    // for FileO, see Edif::Save
    void Load_08( FILE *F_in, string flong );
    void Load_01( FILE *F_in, string flong );
    void LoadScan_01( FILE *F_in, CScan* sout, int nz );
    void Load_96( FILE *F_in, string flong );
}

//! Load file in current or legacy format.
void NFile::Load(void)
{
    string fnames = wask( "Load file(s)" ); // PROVIS, sask sobald glob besser
    vector<string> vflong; // unix names
    if( mystd::glob_file_list( fnames, "", "y08", &vflong )<=0 )
        throw "files " + fnames + " not found";
    FILE *F_in;
    string fdir, fshort, fext;
    // cout << "DEBUG: size " << vflong.size() << "\n";
    NOlm::SelNew();
    for( uint i=0; i<vflong.size(); ++i ) {
        if( !(F_in = fopen(vflong[i].c_str(), "r")) )
            throw string( "cannot open file " ) + vflong[i];
        try{ 
            cout << ".. loading file " << vflong[i] << "\n";
            mystd::fname_divide( vflong[i], &fdir, &fshort, &fext);
            
            if        ( fext=="y08" ) {
                NFileIO::Load_08( F_in, vflong[i] );
            } else if ( fext=="a01" ) {
                NFileIO::Load_01( F_in, vflong[i] );
            } else if ( fext=="i96" ) {
                NFileIO::Load_96( F_in, vflong[i] );
            } else {
                throw "unknown file extension [" + fext + "]";
            }
            
            fclose( F_in );
        }
        catch(exception& e) {
            cout << "DEBUG: catched locally\n";
            cerr << e.what() << endl;
            if (F_in) fclose(F_in);
        }
        catch(string& s) {
            // for y08 only
            cerr << s << endl;
            if (F_in) fclose(F_in);
        }
    }
}

//! Load a YAML file in y08 format.
void NFileIO::Load_08( FILE *F_in, string flong )
{
    string lin, key, val, fdir, fshort, fext;
    double num;
    uint iz;
    COld old;
    CCoord co;
    CScan sout;
    CYamlRead y( F_in );

    mystd::fname_divide( flong, &fdir, &fshort, &fext);

    y.checktype( YAML_STREAM_START_EVENT, "no stream\n" );
    y.checktype( YAML_DOCUMENT_START_EVENT, "no document\n" );
    y.checktype( YAML_MAPPING_START_EVENT, "main: no map\n" );

    y.checkvalue( "Meta" );
    y.checktype( YAML_MAPPING_START_EVENT, "Meta: no map\n" );
    y.checkvalue( "format" );
    y.checkvalue( "frida/y08 for yaml1" );
    y.skipvalues( 2 );
    y.checktype( YAML_MAPPING_END_EVENT, "Meta: no end of map\n" );

    y.checkvalue( "History" );
    y.checktype( YAML_SEQUENCE_START_EVENT, "History: no seq\n" );
    while( y.scalar_from_seq( "history", &lin ) ){
        old.lDoc.push_back(lin);
    }

    y.checkvalue( "Coord" );
    y.checktype( YAML_MAPPING_START_EVENT, "Coord: no map\n" );
    if( !y.getCoord( "x", &(old.xco) ) )
        throw string( "no x coord" );
    if( !y.getCoord( "y", &(old.yco) ) )
        throw string( "no y coord" );
    iz=0;
    while( y.getCoord( "z"+strg(iz), &(co) ) ){
        old.ZCo.push_back( co );
        ++iz;
    }
    if( y.event.type != YAML_MAPPING_END_EVENT )
        throw string( "Coord: no end of map" );
    y.done();

    y.checkvalue( "Param" );
    // Param section may be empty
    y.next();
    if( y.event.type==YAML_SEQUENCE_START_EVENT ){
        y.done();
        while( true ){
            y.next();
            if( y.event.type == YAML_SEQUENCE_END_EVENT )
                break;
            if( y.event.type != YAML_MAPPING_START_EVENT )
                throw string ( "Param: no map" );
            y.done();
            y.checkvalue( "name" );
            co.name = y.getscalar( "param.name" );
            y.checkvalue( "unit" );
            co.unit = y.getscalar( "param(" + co.name + ").unit" );
            y.checkvalue( "value" );
            val = y.getscalar( "param(" + co.name + ").value" );
            if( mystd::any2dbl( val, &num ) )
                throw string( "param(" ) + co.name + "): invalid value " + val;
            old.RPar.push_back( CParam( co, num ) );
            y.checktype( YAML_MAPPING_END_EVENT,
                         "Param "+co.name+": no eom\n" );
        }
    } else {
        if( !y.event.type==YAML_SCALAR_EVENT )
            throw( "Param: no empty scalar in empty param section" );
    }
    y.done();

    y.checkvalue( "Tables" );
    y.checktype( YAML_SEQUENCE_START_EVENT, "Tables: no seq\n" );
    while( true ){
        y.next();
        if( y.event.type == YAML_SEQUENCE_END_EVENT )
            break;
        if( y.event.type != YAML_MAPPING_START_EVENT )
            throw string ( "Tables: no map" );
        y.done();
        sout.Clear();
        for( iz=0; iz<old.ZCo.size(); ++iz ){
            y.checkvalue( "z" + strg(iz) );
            val = y.getscalar( "z" + strg(iz) + ".value" );
            if( mystd::any2dbl( val, &num ) )
                throw "z" + strg(iz) + ": invalid value " + val;
            sout.z.push_back( num );
        }
        y.checkvalue( "xy" );
        val = y.getscalar( "xy core data" );
        double v,w;
        string res;
        while( val != "" ) {
            mystd::string_extract_line( val, &lin, &res );
            if( sscanf( lin.c_str(), "%lg %lg\n", &v, &w )!=2 )
                throw "y08: bad data line: " + lin;
            sout.push_back(v, w);
            val = res;
        };
        old.VS.push_back( sout );
        y.checktype( YAML_MAPPING_END_EVENT, "Table: no eom\n" );
    }
    y.done();

    old.name = fshort;
    old.lDoc.push_back( "loaded as " + flong );
    old.as_on_disk = true;
    NOlm::OloAdd(&old);
}

//! Load a .a01 formatted file.
void NFileIO::Load_01( FILE *F_in, string flong )
{
    string lin, key, val, fdir, fshort, fext;
    uint iz, nj=0, j;
    COld old;
    CCoord co;
    CScan sout;

    mystd::fname_divide( flong, &fdir, &fshort, &fext);

    if     ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "file empty or starting with empty line" );
    if     ( lin!=string("ASCII-01 JWu") )
        throw ".a01 file starts with illegal header [" + lin + "]";

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<5 || 
         lin.substr(0,4)!=string("typ ") )
        throw string( "no valid typ line" );
    if        ( lin.substr(4)=="D" ) {
        // ok
    } else if ( lin.substr(4)=="C" ) {
        throw string( "reading curve files not implemented" );
    } else {
        throw string( "invalid type [" + lin.substr(4) + "]" );
    }

    while (1) { // loop over input lines

        if        ( mystd::freadln(F_in, &lin)<0 )
            throw string( "unexpected end-of-file" );

        // regular exit when "eof" is reached:
        if        (lin=="eof") {
            old.as_on_disk = true;
            NOlm::OloAdd(&old);
            return;
        }

        if        ( lin.size() < 4 )
            throw string( "incomplete line" );

        key = lin.substr(0,4);
        val = lin.substr(4);

        if        (key=="fil ") {
            if (val!=fshort) {
                cout << "warning: file '" << fshort <<
                    "' renamed from '" << val << "'\n";
            }
            old.name = fshort;
        } else if (key=="doc ") {
            old.lDoc.push_back(val);
        } else if (key=="x   ") {
            old.xco.load_a01(val);
        } else if (key=="y   ") {
            old.yco.load_a01(val);
        } else if (lin[0]=='z') {
            if ( sscanf(lin.substr(1,3).c_str(), "%u", &iz)!=1 )
                throw string( "z coordinates must be numbered" );
            if ( iz!=old.ZCo.size() )
                throw string( "z coordinates must be sorted" );
            co.load_a01(val);
            old.ZCo.push_back(co);
        } else if (lin.substr(0,4)=="rpa ") {
            old.RPar.push_back( CParam() );
            if( old.RPar.back().init_from_a01( lin.substr(4) ) )
                throw string( "invalid rpa line\n" );
        } else if (key=="#j  ") {
            if ( sscanf(val.c_str(), "%u", &nj)!=1 )
               throw string( "after key #j no no. of scans\n");
        } else if (key=="scn ") {
            if ( sscanf(val.c_str(), "%u", &j)!=1 ) {
                cout << "warning: number of scan not given\n";
            } else if ( j!=old.nScan() ) {
                cout << "warning: scan " << old.nScan() <<
                    " has label " << j << "\n";
            }
            LoadScan_01( F_in, &sout, old.nZ() );
            old.VS.push_back( sout );
        } else {
            throw "Unexpected key [" + key + "]";
        }
    }
}

//! Load a scan from a .a01 formatted files.
void NFileIO::LoadScan_01( FILE *F_in, CScan* sout, int nz )
{
    int err;
    uint i, m, n;
    char lab[4];
    double v, w;

    sout->Clear();

    // get z values:
    if ((err=fscanf(F_in, "%4c%u\n", lab, &m))!=2)
        throw "expecting Z head line, found " + strg(err) + " entries";
    if (strncmp(lab, "Z   ", 4))
        throw "found [" + strg(lab) + "] in Z head line";
    if (m!=nz)
        throw "found " + strg(m) + " Z entries, expecting " + strg(nz);
    for ( i=0; i<nz; ++i ) {
        if ((err=fscanf(F_in, "%lg\n", &v))!=1)
            throw "expecting z value, found " + strg(err) + " entries";
        sout->z.push_back( v );
    }

    // get xy values:
    if ( (err=fscanf(F_in, "%4c%u\n", lab, &n))!=2 )
        throw "expecting XY head line, found " + strg(err) + " entries";
    if ( strncmp(lab, "XY  ", 4) )
        throw "found [" + strg(lab) + "] in XY head line";
    
    for (i=0; i<n; i++) {
        if ((err=fscanf(F_in, "%lg %lg\n", &v, &w))!=2) {
            throw string( "expecting xy pair, found strg(err) entries" );
        }
        sout->push_back(v, w);
    }
}


//! Load a .i96 formatted file.
void NFileIO::Load_96( FILE *F_in, string flong )
{
    string lin, key, val, fdir, fshort, fext;
    double r0, r1, r2;
    uint n, nspec, ival;
    COld old;
    CCoord co;
    CScan sout;

    mystd::fname_divide( flong, &fdir, &fshort, &fext);
    old.name = fshort;

    if     ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "file empty or starting with empty line" );
    if     ( lin!=string("ASCII-96") )
        throw ".i96 file starts with illegal header [" + lin + "]";

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<9 || 
         lin.substr(0,9)!=string("(a24,a56)") )
        throw string( "no valid intro to block 2" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<24 ||
         lin.substr(0,3)!=string("fil") )
        throw string( "tag 'fil' not found in block 2" );
    if ( mystd::strip(lin.substr(24))!=fshort ) {
                cout << "warning: i96 file '" << fshort <<
                    "' renamed from '" << lin.substr(24) << "'\n";
    }

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<24 ||
         lin.substr(0,3)!=string("tit") )
        throw string( "tag 'tit' not found in block 2" );
    old.lDoc.push_back( mystd::strip(lin.substr(24)) );

    while( mystd::freadln(F_in, &lin)>0 && lin.length()>=24 &&
         lin.substr(0,3)==string("doc") )
        old.lDoc.push_back( mystd::strip(lin.substr(24)) );

    if( lin.substr(0,3)!=string("dir") )
        throw string( "tag 'dir' not found in block 2" );

    while ( mystd::freadln(F_in, &lin)>0 &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 2") ) )
        ;
    
    if ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "missed &eob 3" );

    if( lin.length()<9 || lin.substr(0,9)!=string("(a24,i16)") )
        throw string( "no valid intro to block 3" );

    while ( mystd::freadln(F_in, &lin)>0 &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 3") ) )
        ;
    
    if ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "missed &eob 3" );

    if ( lin.length()<16 || lin.substr(0,16)!=string("(a24,a24,g20.10)") )
        throw string( "no valid intro to block 4" );

    while ( mystd::freadln(F_in, &lin)>0 && lin.length()>=48 &&
            lin.substr(0,4)!=string("&eob") ){
        if( sscanf( lin.substr(48).c_str(), "%lg", &r0 )!=1 )
            throw string( "no real value in rpar line" );
        old.RPar.push_back(
            CParam( lin.substr(0,24), lin.substr(24,24), r0 ) );
    }

    if( lin.substr(0,6)!=string("&eob 4") )
        throw string( "no valid eob 4" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<9 || 
         lin.substr(0,12)!=string("(a4,a24,a24)") )
        throw string( "no valid intro to block 5" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<28 || 
         lin.substr(0,1)!=string("x") )
        throw string( "no x coordinate" );
    old.xco = CCoord( mystd::strip(lin.substr(4,24)), mystd::strip(lin.substr(28,24)) );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<28 || 
         lin.substr(0,1)!=string("y") )
        throw string( "no y coordinate" );
    old.yco = CCoord( mystd::strip(lin.substr(4,24)), mystd::strip(lin.substr(28,24)) );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<28 || 
         lin.substr(0,1)!=string("z") )
        throw string( "no z coordinate" );
    old.ZCo.push_back( CCoord( mystd::strip(lin.substr(4,24)), mystd::strip(lin.substr(28,24)) ) );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<6 || 
         lin.substr(0,6)!=string("&eob 5") )
        throw string( "no valid eob 5" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<5 || 
         lin.substr(0,5)!=string("(a80)") )
        throw string( "no valid intro to block 6" );

    while ( mystd::freadln(F_in, &lin)>0 &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 6") ) )
        old.lDoc.push_back( lin );

    if ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "missed &eob 6" );
    for( int i=1; i<3; ++i )
        if ( mystd::freadln(F_in, &lin)<=0 )
            throw string( "no valid intro to block 7" );

    if ( mystd::freadln(F_in, &lin)<=0 || 
         sscanf( lin.c_str(), "%i %i", &nspec, &ival )!=2 )
            throw string( "no nspec or 2nd arg in intro to block 7" );

    for( size_t j=0; j<nspec; ++j ){

        sout.Clear();

        if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<10 || 
             lin.substr(0,9)!=string("&spectrum") ||
             sscanf( lin.substr(10).c_str(), "%i", &ival )!=1 ||
             ival!=j+1 )
            throw string( "no valid intro to spectrum" );

        if ( mystd::freadln(F_in, &lin)<0 ||
             sscanf( lin.c_str(), "%i %lg", &n, &r0 )!=2 )
            throw string( "no valid header in spectrum" );

        sout.z.push_back( r0 );

        for( size_t i=0; i<n; ++i ){
            if ( !F_in || mystd::freadln(F_in, &lin)<=0 )
                throw "i96: failed to read data line " +
                    strg(i) + " of " + strg(n);
            if ( sscanf( lin.c_str(), "%lg %lg %lg", &r0, &r1, &r2 )!=3 )
                throw "i96: bad data line: '" + lin + "'";
            sout.push_back(r0, r1);
        }

        old.VS.push_back( sout );

    }

    old.as_on_disk = true;
    NOlm::OloAdd(&old);

}
